package java.util.stream;

import java.util.function.DoubleConsumer;
import java.util.function.Consumer;
import java.util.function.LongConsumer;
import java.util.Comparator;
import java.util.Objects;
import java.util.function.IntConsumer;
import java.util.Spliterator;

final class Streams
{
    static final Object NONE;
    
    private Streams() {
        throw new Error("no instances");
    }
    
    static Runnable composeWithExceptions(final Runnable runnable, final Runnable runnable2) {
        return new Runnable() {
            @Override
            public void run() {
                try {
                    runnable.run();
                }
                catch (Throwable t) {
                    try {
                        runnable2.run();
                    }
                    catch (Throwable exception) {
                        try {
                            t.addSuppressed(exception);
                        }
                        catch (Throwable t2) {}
                    }
                    throw t;
                }
                runnable2.run();
            }
        };
    }
    
    static Runnable composedClose(final BaseStream<?, ?> baseStream, final BaseStream<?, ?> baseStream2) {
        return new Runnable() {
            @Override
            public void run() {
                try {
                    baseStream.close();
                }
                catch (Throwable t) {
                    try {
                        baseStream2.close();
                    }
                    catch (Throwable exception) {
                        try {
                            t.addSuppressed(exception);
                        }
                        catch (Throwable t2) {}
                    }
                    throw t;
                }
                baseStream2.close();
            }
        };
    }
    
    static {
        NONE = new Object();
    }
    
    static final class RangeIntSpliterator implements Spliterator.OfInt
    {
        private int from;
        private final int upTo;
        private int last;
        private static final int BALANCED_SPLIT_THRESHOLD = 16777216;
        private static final int RIGHT_BALANCED_SPLIT_RATIO = 8;
        
        RangeIntSpliterator(int n, int n2, boolean b) {
            this(n, n2, b ? 1 : 0);
        }
        
        private RangeIntSpliterator(int from, int upTo, int last) {
            this.from = from;
            this.upTo = upTo;
            this.last = last;
        }
        
        @Override
        public boolean tryAdvance(IntConsumer obj) {
            Objects.requireNonNull(obj);
            int from = this.from;
            if (from < this.upTo) {
                ++this.from;
                obj.accept(from);
                return true;
            }
            if (this.last > 0) {
                this.last = 0;
                obj.accept(from);
                return true;
            }
            return false;
        }
        
        @Override
        public void forEachRemaining(IntConsumer obj) {
            Objects.requireNonNull(obj);
            int i = this.from;
            int upTo = this.upTo;
            int last = this.last;
            this.from = this.upTo;
            this.last = 0;
            while (i < upTo) {
                obj.accept(i++);
            }
            if (last > 0) {
                obj.accept(i);
            }
        }
        
        @Override
        public long estimateSize() {
            return (long)this.upTo - (long)this.from + (long)this.last;
        }
        
        @Override
        public int characteristics() {
            return 17749;
        }
        
        @Override
        public Comparator<? super Integer> getComparator() {
            return null;
        }
        
        @Override
        public Spliterator.OfInt trySplit() {
            long estimateSize = this.estimateSize();
            return (estimateSize <= 1L) ? null : new Streams.RangeIntSpliterator(this.from, this.from += this.splitPoint(estimateSize), 0);
        }
        
        private int splitPoint(long n) {
            return (int)(n / (long)((n < 16777216L) ? 2 : 8));
        }
    }
    
    static final class RangeLongSpliterator implements Spliterator.OfLong
    {
        private long from;
        private final long upTo;
        private int last;
        private static final long BALANCED_SPLIT_THRESHOLD = 16777216L;
        private static final long RIGHT_BALANCED_SPLIT_RATIO = 8L;
        
        RangeLongSpliterator(long n, long n2, boolean b) {
            this(n, n2, b ? 1 : 0);
        }
        
        private RangeLongSpliterator(long from, long upTo, int last) {
            assert upTo - from + (long)last > 0L;
            this.from = from;
            this.upTo = upTo;
            this.last = last;
            return;
        }
        
        @Override
        public boolean tryAdvance(LongConsumer obj) {
            Objects.requireNonNull(obj);
            long from = this.from;
            if (from < this.upTo) {
                ++this.from;
                obj.accept(from);
                return true;
            }
            if (this.last > 0) {
                this.last = 0;
                obj.accept(from);
                return true;
            }
            return false;
        }
        
        @Override
        public void forEachRemaining(LongConsumer obj) {
            Objects.requireNonNull(obj);
            long from = this.from;
            long upTo = this.upTo;
            int last = this.last;
            this.from = this.upTo;
            this.last = 0;
            while (from < upTo) {
                obj.accept(from++);
            }
            if (last > 0) {
                obj.accept(from);
            }
        }
        
        @Override
        public long estimateSize() {
            return this.upTo - this.from + (long)this.last;
        }
        
        @Override
        public int characteristics() {
            return 17749;
        }
        
        @Override
        public Comparator<? super Long> getComparator() {
            return null;
        }
        
        @Override
        public Spliterator.OfLong trySplit() {
            long estimateSize = this.estimateSize();
            return (estimateSize <= 1L) ? null : new Streams.RangeLongSpliterator(this.from, this.from += this.splitPoint(estimateSize), 0);
        }
        
        private long splitPoint(long n) {
            return n / ((n < 16777216L) ? 2L : 8L);
        }
    }
    
    private abstract static class AbstractStreamBuilderImpl<T, S extends Spliterator<T>> implements Spliterator<T>
    {
        int count;
        
        @Override
        public S trySplit() {
            return null;
        }
        
        @Override
        public long estimateSize() {
            return (long)(-this.count - 1);
        }
        
        @Override
        public int characteristics() {
            return 17488;
        }
    }
    
    static final class StreamBuilderImpl<T> extends Streams.AbstractStreamBuilderImpl<T, Spliterator<T>> implements Stream.Builder<T>
    {
        T first;
        SpinedBuffer<T> buffer;
        
        StreamBuilderImpl() {
        }
        
        StreamBuilderImpl(T first) {
            this.first = first;
            this.count = -2;
        }
        
        @Override
        public void accept(T first) {
            if (this.count == 0) {
                this.first = first;
                ++this.count;
            }
            else if (this.count > 0) {
                if (this.buffer == null) {
                    (this.buffer = new SpinedBuffer()).accept(this.first);
                    ++this.count;
                }
                this.buffer.accept(first);
            }
            else {
                throw new IllegalStateException();
            }
        }
        
        @Override
        public Stream.Builder<T> add(T t) {
            this.accept(t);
            return this;
        }
        
        @Override
        public Stream<T> build() {
            int count = this.count;
            if (count >= 0) {
                this.count = -this.count - 1;
                return (count < 2) ? StreamSupport.stream(this, false) : StreamSupport.stream(this.buffer.spliterator(), false);
            }
            throw new IllegalStateException();
        }
        
        @Override
        public boolean tryAdvance(Consumer<? super T> obj) {
            Objects.requireNonNull(obj);
            if (this.count == -2) {
                obj.accept(this.first);
                this.count = -1;
                return true;
            }
            return false;
        }
        
        @Override
        public void forEachRemaining(Consumer<? super T> obj) {
            Objects.requireNonNull(obj);
            if (this.count == -2) {
                obj.accept(this.first);
                this.count = -1;
            }
        }
    }
    
    static final class IntStreamBuilderImpl extends Streams.AbstractStreamBuilderImpl<Integer, Spliterator.OfInt> implements IntStream.Builder, Spliterator.OfInt
    {
        int first;
        SpinedBuffer.OfInt buffer;
        
        IntStreamBuilderImpl() {
        }
        
        IntStreamBuilderImpl(int first) {
            this.first = first;
            this.count = -2;
        }
        
        @Override
        public void accept(int first) {
            if (this.count == 0) {
                this.first = first;
                ++this.count;
            }
            else if (this.count > 0) {
                if (this.buffer == null) {
                    (this.buffer = new SpinedBuffer.OfInt()).accept(this.first);
                    ++this.count;
                }
                this.buffer.accept(first);
            }
            else {
                throw new IllegalStateException();
            }
        }
        
        @Override
        public IntStream build() {
            int count = this.count;
            if (count >= 0) {
                this.count = -this.count - 1;
                return (count < 2) ? StreamSupport.intStream(this, false) : StreamSupport.intStream(this.buffer.spliterator(), false);
            }
            throw new IllegalStateException();
        }
        
        @Override
        public boolean tryAdvance(IntConsumer obj) {
            Objects.requireNonNull(obj);
            if (this.count == -2) {
                obj.accept(this.first);
                this.count = -1;
                return true;
            }
            return false;
        }
        
        @Override
        public void forEachRemaining(IntConsumer obj) {
            Objects.requireNonNull(obj);
            if (this.count == -2) {
                obj.accept(this.first);
                this.count = -1;
            }
        }
    }
    
    static final class LongStreamBuilderImpl extends Streams.AbstractStreamBuilderImpl<Long, Spliterator.OfLong> implements LongStream.Builder, Spliterator.OfLong
    {
        long first;
        SpinedBuffer.OfLong buffer;
        
        LongStreamBuilderImpl() {
        }
        
        LongStreamBuilderImpl(long first) {
            this.first = first;
            this.count = -2;
        }
        
        @Override
        public void accept(long first) {
            if (this.count == 0) {
                this.first = first;
                ++this.count;
            }
            else if (this.count > 0) {
                if (this.buffer == null) {
                    (this.buffer = new SpinedBuffer.OfLong()).accept(this.first);
                    ++this.count;
                }
                this.buffer.accept(first);
            }
            else {
                throw new IllegalStateException();
            }
        }
        
        @Override
        public LongStream build() {
            int count = this.count;
            if (count >= 0) {
                this.count = -this.count - 1;
                return (count < 2) ? StreamSupport.longStream(this, false) : StreamSupport.longStream(this.buffer.spliterator(), false);
            }
            throw new IllegalStateException();
        }
        
        @Override
        public boolean tryAdvance(LongConsumer obj) {
            Objects.requireNonNull(obj);
            if (this.count == -2) {
                obj.accept(this.first);
                this.count = -1;
                return true;
            }
            return false;
        }
        
        @Override
        public void forEachRemaining(LongConsumer obj) {
            Objects.requireNonNull(obj);
            if (this.count == -2) {
                obj.accept(this.first);
                this.count = -1;
            }
        }
    }
    
    static final class DoubleStreamBuilderImpl extends Streams.AbstractStreamBuilderImpl<Double, Spliterator.OfDouble> implements DoubleStream.Builder, Spliterator.OfDouble
    {
        double first;
        SpinedBuffer.OfDouble buffer;
        
        DoubleStreamBuilderImpl() {
        }
        
        DoubleStreamBuilderImpl(double first) {
            this.first = first;
            this.count = -2;
        }
        
        @Override
        public void accept(double first) {
            if (this.count == 0) {
                this.first = first;
                ++this.count;
            }
            else if (this.count > 0) {
                if (this.buffer == null) {
                    (this.buffer = new SpinedBuffer.OfDouble()).accept(this.first);
                    ++this.count;
                }
                this.buffer.accept(first);
            }
            else {
                throw new IllegalStateException();
            }
        }
        
        @Override
        public DoubleStream build() {
            int count = this.count;
            if (count >= 0) {
                this.count = -this.count - 1;
                return (count < 2) ? StreamSupport.doubleStream(this, false) : StreamSupport.doubleStream(this.buffer.spliterator(), false);
            }
            throw new IllegalStateException();
        }
        
        @Override
        public boolean tryAdvance(DoubleConsumer obj) {
            Objects.requireNonNull(obj);
            if (this.count == -2) {
                obj.accept(this.first);
                this.count = -1;
                return true;
            }
            return false;
        }
        
        @Override
        public void forEachRemaining(DoubleConsumer obj) {
            Objects.requireNonNull(obj);
            if (this.count == -2) {
                obj.accept(this.first);
                this.count = -1;
            }
        }
    }
    
    abstract static class ConcatSpliterator<T, T_SPLITR extends Spliterator<T>> implements Spliterator<T>
    {
        protected final T_SPLITR aSpliterator;
        protected final T_SPLITR bSpliterator;
        boolean beforeSplit;
        final boolean unsized;
        
        public ConcatSpliterator(T_SPLITR aSpliterator, T_SPLITR bSpliterator) {
            this.aSpliterator = aSpliterator;
            this.bSpliterator = bSpliterator;
            this.beforeSplit = true;
            this.unsized = (aSpliterator.estimateSize() + bSpliterator.estimateSize() < 0L);
        }
        
        @Override
        public T_SPLITR trySplit() {
            Spliterator<Object> spliterator = this.beforeSplit ? this.aSpliterator : this.bSpliterator.trySplit();
            this.beforeSplit = false;
            return spliterator;
        }
        
        @Override
        public boolean tryAdvance(Consumer<? super T> consumer) {
            int n;
            if (this.beforeSplit) {
                n = this.aSpliterator.tryAdvance(consumer);
                if (n == 0) {
                    this.beforeSplit = false;
                    n = this.bSpliterator.tryAdvance(consumer);
                }
            }
            else {
                n = this.bSpliterator.tryAdvance(consumer);
            }
            return n;
        }
        
        @Override
        public void forEachRemaining(Consumer<? super T> consumer) {
            if (this.beforeSplit) {
                this.aSpliterator.forEachRemaining(consumer);
            }
            this.bSpliterator.forEachRemaining(consumer);
        }
        
        @Override
        public long estimateSize() {
            if (this.beforeSplit) {
                long n = this.aSpliterator.estimateSize() + this.bSpliterator.estimateSize();
                return (n >= 0L) ? n : Long.MAX_VALUE;
            }
            return this.bSpliterator.estimateSize();
        }
        
        @Override
        public int characteristics() {
            if (this.beforeSplit) {
                return this.aSpliterator.characteristics() & this.bSpliterator.characteristics() & ~(0x5 | (this.unsized ? 16448 : 0));
            }
            return this.bSpliterator.characteristics();
        }
        
        @Override
        public Comparator<? super T> getComparator() {
            if (this.beforeSplit) {
                throw new IllegalStateException();
            }
            return this.bSpliterator.getComparator();
        }
        
        static class OfRef<T> extends Streams.ConcatSpliterator<T, Spliterator<T>>
        {
            OfRef(Spliterator<T> spliterator, Spliterator<T> spliterator2) {
                super(spliterator, spliterator2);
            }
        }
        
        private abstract static class OfPrimitive<T, T_CONS, T_SPLITR extends Spliterator.OfPrimitive<T, T_CONS, T_SPLITR>> extends Streams.ConcatSpliterator<T, T_SPLITR> implements Spliterator.OfPrimitive<T, T_CONS, T_SPLITR>
        {
            private OfPrimitive(T_SPLITR t_SPLITR, T_SPLITR t_SPLITR2) {
                super(t_SPLITR, t_SPLITR2);
            }
            
            @Override
            public boolean tryAdvance(T_CONS t_CONS) {
                int n;
                if (this.beforeSplit) {
                    n = ((Spliterator.OfPrimitive)this.aSpliterator).tryAdvance(t_CONS);
                    if (n == 0) {
                        this.beforeSplit = false;
                        n = ((Spliterator.OfPrimitive)this.bSpliterator).tryAdvance(t_CONS);
                    }
                }
                else {
                    n = ((Spliterator.OfPrimitive)this.bSpliterator).tryAdvance(t_CONS);
                }
                return n;
            }
            
            @Override
            public void forEachRemaining(T_CONS t_CONS) {
                if (this.beforeSplit) {
                    ((Spliterator.OfPrimitive)this.aSpliterator).forEachRemaining(t_CONS);
                }
                ((Spliterator.OfPrimitive)this.bSpliterator).forEachRemaining(t_CONS);
            }
        }
        
        static class OfInt extends Streams.ConcatSpliterator.OfPrimitive<Integer, IntConsumer, Spliterator.OfInt> implements Spliterator.OfInt
        {
            OfInt(Spliterator.OfInt ofInt, Spliterator.OfInt ofInt2) {
                super(ofInt, ofInt2);
            }
        }
        
        static class OfLong extends Streams.ConcatSpliterator.OfPrimitive<Long, LongConsumer, Spliterator.OfLong> implements Spliterator.OfLong
        {
            OfLong(Spliterator.OfLong ofLong, Spliterator.OfLong ofLong2) {
                super(ofLong, ofLong2);
            }
        }
        
        static class OfDouble extends Streams.ConcatSpliterator.OfPrimitive<Double, DoubleConsumer, Spliterator.OfDouble> implements Spliterator.OfDouble
        {
            OfDouble(Spliterator.OfDouble ofDouble, Spliterator.OfDouble ofDouble2) {
                super(ofDouble, ofDouble2);
            }
        }
    }
}
